﻿// ***** TEST HARNESS *****
// This contains a private key that, at the time of writing, matches the DKIM public key for dev.citadel.org
// We can use the attached test message to validate a signature against that.

#include <stdio.h>
#include <syslog.h>
#include <libcitadel.h>

// This was easier than trying to figure out the header situation
void dkim_sign(StrBuf *email, char *pkey_in, char *domain, char *selector);

int main(int argc, char *argv[]) {

	// display the greeting
	fprintf(stderr,
		"\033[44m\033[1m╔════════════════════════════════════════════════════════════════════════╗\033[0m\n"
		"\033[44m\033[1m║ DKIM signature test program for Citadel                                ║\033[0m\n"
		"\033[44m\033[1m║ Copyright (c) 2024 by citadel.org (Art Cancro et al.)                  ║\033[0m\n"
		"\033[44m\033[1m║ This program is open source software.  Use, duplication, or disclosure ║\033[0m\n"
		"\033[44m\033[1m║ is subject to the terms of the GNU General Public license v3.          ║\033[0m\n"
		"\033[44m\033[1m╚════════════════════════════════════════════════════════════════════════╝\033[0m\n"
	);

	openlog("dkim", LOG_PERROR, LOG_USER);

	// dkim.c can handle a PEM-encoded PKCS#7 private key that has had all of the newlines replaced by underscores.
	// It will convert them back to newlines before importing the key.
	// This is the format that Citadel Server uses to store the key in its configuration database.
	char *private_key = "-----BEGIN PRIVATE KEY-----_MIIEvAIBADANBgkqhkiG9w0BAQEFAASCBKYwggSiAgEAAoIBAQDfuefcepokRrnp_SSDsxu+QDqeD8GL9QnZz/N6IxTdBv6Wc10ExBe2IjS5dKI7AvhSSEK0zGE8Hkpmw_eccbiepQqeueteWzAMZ1uT43bD3k7eye7vWobiOP9QtoYGR6sG25h2W5Tbc91W4f_dvYnxYVJjx8wIVF0f3o25v+rQueoo0HlvGyA9/xi9GAaJL05OmK1xnMJgSvW/Q8Q_zq7apf1D6XPXHuhv5tevElkZ5jlvM2w0cTVyAzMrUh6Rkcn9xM4/NPWYghBc3jO4_TrPnSrobQGrX0fcizE/FN6I0in0Ke8Z+gMM8NeFcsjvLZe9MpY9i0pw/ygLIh5t3_O4qpwC1JAgMBAAECggEAIwiTCMEAGzciDKhhagJ66BWLYMtHTP5X2zDZThSH4xlW_HznL4RfbCtuEy5y6we7h/L90x8ACPB7WRz7CkYrmsMvy9A7q0b2I1k10MyyVgqBJ_QdgMitv4YKYQK7+QbG/tNrS/lqVXUOz3iiDQSgkRpqOtUBWfkj0WD7vbhF99NDhV_dxaehFkKv3yNy0bXJlHJBJ6KtOUnDwub8TExh8dyj3kB8Qzj4I98shaXPNUSSaOw_zG6QG72yrxlMs495jkIPbF2JDidmLrX+oVISwKyaBWx+BkFV/KFAEKgaB5/nCw7+_qq/jxsmXim3HuQ3MIAjq1yw9aGRH1HMi8Gn7tYlNGwKBgQDy6EEKpuEiW9wwlI2+_GVuSkhSTTX1h6qK/ay8Jtyb8yJM/BxogAQlfjdgFixiZHy5MaomTbfeT2GDji553_+RsnZ60+g7FI9nHwabSxtuCQ+vjbFqCsdMPAiSeG0bEzo0zf5TjASdUtuZL0vXjl_yMZWDEuESoVNlYlvCOVkw2nvIwKBgQDryPuSq6PNVHRWsKRRs5ju4wKs/1ucBOg5_gCcN8lE03mFCWAlZhypE4/fAhTQ/a5KQoAzc0QZcXRueDyNsnc+QWw3/QWf8/fkV_HPfTWS3Dcuj+4RnWUucaZ/mKFlTC3+eNSlpyaPIMlCjXGsJ9GlPrsaAi9KPbD2v/_XcMq/PMOowKBgHVf7S3sfZVQthFzdxqIvksQ84hKRW/vJT1B2bTkH56+fQhTsjgM_yC64J85l7DjxbDnYsSngVWXHhOnvKV/nq0tbOcefcydCjsQREBNfvxvPajjTskgj_FAQRQlxPL0U4f4khBk9EXhJ+PZithaHjZpNl1YfTSp62x3Yz4kTSeHnpAoGAGn5m_5kArE7NdrzACBrwrfww7DL1Uyd8zSOLBgKutvEcQnqfNxSWO9la3TAarrESmH2Ic_j+Nc15wOsl/5FwdUf1/73qa2zJKtHlY28qSeo8uRqrIYeSCvnyP3wjBoLc2C8zlb_mGd6azdqr2DuYahHrcAzwjnC/6Zn+DXM7FOn7AkCgYBp1xxY88cCoF24yffkD3MC_ACUury4qRSDTGx6/qCCkIyWxg1vuiDrlPWhSwQznxHvovcfpdjdbWcFY87IK6mpG_aJHwMJ7Kw+baoxGPZWHwdg6BgvUCihe3xlcaq6rOBoLviD6FOzbogg++Tvi0LemG_y/wEs/mZkaRzW4n41ir0Xw==_-----END PRIVATE KEY-----";

	char *domain = "dev.citadel.org";
	char *selector = "foo";

	// Sample message
	StrBuf *email = NewStrBufPlain(HKEY(
		"From: Fred Bloggs <bloggs@dev.citadel.org>\r\n"
		"X-irrelevant-header: wow mom 303\r\n"
		"To: Bread Floggs <bf@example.com>\r\n"
		"Subject: The ultimate  test message\r\n"
		"Message-ID: <73294856-8726543-473298@dev.citadel.org>\r\n"
		"\r\n"
		"Hi.\r\n"
		"\r\n"
		"Bhille Disassemble.  Highly recommend.\r\n"
		"\r\n"
		"--Fred\r\n"


	));

	// create signature
	dkim_sign(email, private_key, domain, selector);

	// Show the user what we did
	printf("%s", (char *)ChrPtr(email));

	FILE *fp;
	printf("\033[34m-----\033[0m\n");
	printf("Piping original version to test program (this should pass)\n");
	fp = popen("./tester.pl | sed s/pass/\033[32mpass\033[0m/g | sed s/fail/\033[31mfail\033[0m/g", "w");
	fwrite((char *)ChrPtr(email), StrLength(email), 1, fp);
	pclose(fp);
	printf("\033[34m-----\033[0m\n");
	printf("Piping altered version to test program (this should fail)\n");
	fp = popen("sed s/oggs/argh/g | ./tester.pl | sed s/pass/\033[32mpass\033[0m/g | sed s/fail/\033[31mfail\033[0m/g", "w");
	fwrite((char *)ChrPtr(email), StrLength(email), 1, fp);
	pclose(fp);
	printf("\033[34m-----\033[0m\n");

	// free some memory
	FreeStrBuf(&email);

	// exit the test program
	return(0);
}
