// This file contains functions which handle all of the CalDAV "REPORT" queries specified in RFC4791 section 7.
// Copyright (c) 2023-2024 by the citadel.org team
// This program is open source software.  Use, duplication, or disclosure is subject to the GNU General Public License v3.

#include "webcit.h"

#define CALDAV "urn:ietf:params:xml:ns:caldav:"		// Shorthand for the XML namespace of CalDAV
#define CALDAVLEN sizeof(CALDAV)-1			// And the length of that string

const char *the_beginning_of_time	= "19010101T010101Z";
const char *the_end_of_time		= "99991231T235959Z";

// A CalDAV REPORT can only be one type.  This is stored in the report_type member.
enum cr_type {
	cr_calendar_query,
	cr_calendar_multiget,
	cr_freebusy_query
};

// Data type for CalDAV Report Parameters.
// As we slog our way through the XML we learn what the client is asking for
// and build up the contents of this data type.
struct cr_params {
	int comp_filter_nesting_level;
	enum cr_type report_type;	// which RFC4791 section 7 REPORT are we generating
	StrBuf *Chardata;		// XML chardata in between tags is built up here
	StrBuf *Hrefs;			// list of items requested by a `calendar-multiget` REPORT
	Array *filters;			// If the query contains a FILTER stanza, the filter criteria are populated here
	int filter_nest;		// tag nesting level where a FILTER stanza begins
	int in_prop;			// nonzero if we are within the parsing a <DAV::prop> stanza?
	int yes_getetag;		// nonzero if client has requested the "getetag" property
	int yes_calendar_data;		// nonzero if client has requested the "calendar-data" property
};


// XML parser callback
void caldav_xml_start(void *data, const char *el, const char **attr) {
	struct cr_params *crp = (struct cr_params *) data;

#ifdef DEBUG_XML_PARSE
	// logging
	syslog(LOG_DEBUG, "<%s>", el);
	for (int i = 0; attr[i] != NULL; i += 2) {
		syslog(LOG_DEBUG, "Attribute '%s' = '%s'", attr[i], attr[i + 1]);
	}
	// end logging
#endif

	// if we are already within a webdav prop stanza, child nodes name properties the client wants to receive.
	if (crp->in_prop) {
		syslog(LOG_DEBUG, "\033[41m\033[37m\033[1mprop:%s\033[0m", el);

			if (!strcasecmp(el, "DAV::getetag")) {
				crp->yes_getetag = 1;
			}
			else if (!strcasecmp(el, CALDAV"calendar-data")) {
				crp->yes_calendar_data = 1;
			}

		return;
	}

	// webdav prop (not caldav "prop") element
	if (!strcasecmp(el, "DAV::prop")) {
		++crp->in_prop;
		syslog(LOG_DEBUG, "\033[41m\033[37m\033[1mprop:%d\033[0m", crp->in_prop);
	}

	// webdav allprop (not caldav "allprop") element
	if (!strcasecmp(el, "DAV::allprop")) {
			crp->yes_getetag = 1;
			crp->yes_calendar_data = 1;
	}

	// RFC4791 7.8 "calendar-query" REPORT - Client will send a lot of search criteria.
	if (!strcasecmp(el, CALDAV"calendar-query")) {
		crp->report_type = cr_calendar_query;
	}

	// RFC4791 7.9 "calendar-multiget" REPORT - Client will supply a list of specific hrefs.
	else if (!strcasecmp(el, CALDAV"calendar-multiget")) {
		crp->report_type = cr_calendar_multiget;
	}

	// RFC4791 7.10 "free-busy-query" REPORT
	else if (!strcasecmp(el, CALDAV"free-busy-query")) {
		crp->report_type = cr_freebusy_query;
	}

	// RFC4791 9.7 create a filter array if this query contains a "filter" stanza
	else if (!strcasecmp(el, CALDAV"filter")) {
		crp->filters = array_new(SIZ);
		crp->filter_nest = crp->comp_filter_nesting_level;
	}

	// Handle the filters defined in RFC4791 9.7.1 through 9.7.5
	else if (	(	(!strcasecmp(el, CALDAV"comp-filter"))
				|| (!strcasecmp(el, CALDAV"prop-filter"))
				|| (!strcasecmp(el, CALDAV"param-filter"))
				|| (!strcasecmp(el, CALDAV"is-not-defined"))
				|| (!strcasecmp(el, CALDAV"text-match"))
				|| (!strcasecmp(el, CALDAV"time-range"))
			)
			&& (crp->filters)			// Make sure we actually allocated an array
	) {

		if (!strcasecmp(el, CALDAV"comp-filter")) {
			++crp->comp_filter_nesting_level;
		}

		char newfilter[SIZ];
		int a = 0;
		int len = snprintf(newfilter, SIZ, "%d|", crp->comp_filter_nesting_level - crp->filter_nest - 1);
		len += snprintf(&newfilter[len], SIZ-len, "%s", &el[CALDAVLEN]);	// filter name without the namespace
		while (attr[a]) {
			len += snprintf(&newfilter[len], SIZ-len, "|%s", attr[a++]);	// now save the attributes
		}
		array_append(crp->filters, newfilter);
	}
}


// XML parser callback
void caldav_xml_end(void *data, const char *el) {
	struct cr_params *crp = (struct cr_params *) data;

#ifdef DEBUG_XML_PARSE
	// logging
	int i;
	syslog(LOG_DEBUG, "</%s>", el);
	// end logging
#endif

	// webdav prop (not caldav "prop") filter
	if (!strcasecmp(el, "DAV::prop")) {
		--crp->in_prop;
		syslog(LOG_DEBUG, "\033[41m\033[37m\033[1mprop:%d\033[0m", crp->in_prop);
	}

	if (!strcasecmp(el, CALDAV"comp-filter")) {
		--crp->comp_filter_nesting_level;
	}

	if ((!strcasecmp(el, "DAV::href")) || (!strcasecmp(el, "DAV:href"))) {
		if (crp->Hrefs == NULL) {	// append crp->Chardata to crp->Hrefs
			crp->Hrefs = NewStrBuf();
		}
		else {
			StrBufAppendBufPlain(crp->Hrefs, HKEY("|"), 0);
		}
		StrBufAppendBuf(crp->Hrefs, crp->Chardata, 0);
	}

	if (crp->Chardata != NULL) {		// Tag is closed; chardata is now out of scope.
		FreeStrBuf(&crp->Chardata);	// Free the buffer.
		crp->Chardata = NULL;
	}
}


// XML parser callback
void caldav_xml_chardata(void *data, const XML_Char *s, int len) {
	struct cr_params *crp = (struct cr_params *) data;

	char *app = malloc(len+1);
	if (!app) {
		return;
	}
	memcpy(app, s, len);
	app[len] = 0;

	if (crp->Chardata == NULL) {
		crp->Chardata = NewStrBuf();
	}

	StrBufAppendBufPlain(crp->Chardata, app, len, 0);

#ifdef DEBUG_XML_PARSE
	// logging
	string_trim(app);		// remove leading/trailing whitespace.  ok to mangle it because we've already appended.
	if (!IsEmptyStr(app)) {
		int i;
		syslog(LOG_DEBUG, "%s", app, len);
	}
	// end logging
#endif

	free(app);
	return;
}


// Called by caldav_report_one_item() to fetch a message (by number) in the current room,
// and return only the icalendar data as a StrBuf.  Returns NULL if not found.
//
// NOTE: this function expects that "MSGP text/calendar" was issued at the beginning
// of a REPORT operation to set our preferred MIME type to calendar data.
StrBuf *fetch_ical(struct ctdlsession *c, long msgnum) {
	char buf[1024];
	StrBuf *Buf = NULL;

	ctdl_printf(c, "MSG4 %ld", msgnum);
	ctdl_readline(c, buf, sizeof(buf));
	if (buf[0] != '1') {
		return NULL;
	}

	while (ctdl_readline(c, buf, sizeof(buf)), strcmp(buf, "000")) {
		if (Buf != NULL) {		// already in body
			StrBufAppendPrintf(Buf, "%s\n", buf);
		}
		else if (IsEmptyStr(buf)) {	// beginning of body
			Buf = NewStrBuf();
		}
	}

	return Buf;
}


// Called by multiple REPORT types to actually perform the output in "multiget" format.
// We need to already know the source message number and the href, but also already have the output data.
void cal_multiget_out(long msgnum, StrBuf *ThisHref, StrBuf *Caldata, StrBuf *ReportOut, struct cr_params *crp) {

	StrBufAppendPrintf(ReportOut, "<D:response>");
	StrBufAppendPrintf(ReportOut, "<D:href>");
	StrBufXMLEscAppend(ReportOut, ThisHref, NULL, 0, 0);
	StrBufAppendPrintf(ReportOut, "</D:href>");
	StrBufAppendPrintf(ReportOut, "<D:propstat>");

	if (Caldata != NULL) {
		// syslog(LOG_DEBUG, "caldav_report_one_item(%s) 200 OK", ChrPtr(ThisHref));
		StrBufAppendPrintf(ReportOut, "<D:status>");
		StrBufAppendPrintf(ReportOut, "HTTP/1.1 200 OK");
		StrBufAppendPrintf(ReportOut, "</D:status>");
		StrBufAppendPrintf(ReportOut, "<D:prop>");

		if (crp->yes_getetag) {
			StrBufAppendPrintf(ReportOut, "<D:getetag>");
			StrBufAppendPrintf(ReportOut, "%ld", msgnum);
			StrBufAppendPrintf(ReportOut, "</D:getetag>");
		}

		if (crp->yes_calendar_data) {
			StrBufAppendPrintf(ReportOut, "<C:calendar-data>");
			StrBufXMLEscAppend(ReportOut, Caldata, NULL, 0, 0);
			StrBufAppendPrintf(ReportOut, "</C:calendar-data>");
		}

		StrBufAppendPrintf(ReportOut, "</D:prop>");
	}
	else {
		// syslog(LOG_DEBUG, "caldav_report_one_item(%s) 404 not found", ChrPtr(ThisHref));
		StrBufAppendPrintf(ReportOut, "<D:status>");
		StrBufAppendPrintf(ReportOut, "HTTP/1.1 404 not found");
		StrBufAppendPrintf(ReportOut, "</D:status>");
	}

	StrBufAppendPrintf(ReportOut, "</D:propstat>");
	StrBufAppendPrintf(ReportOut, "</D:response>");
}


// Called by caldav_report() to output a single item.
// Our policy is to throw away the list of properties the client asked for, and just send everything.
void caldav_report_one_item(struct http_transaction *h, struct ctdlsession *c, StrBuf *ReportOut, StrBuf *ThisHref, struct cr_params *crp) {
	long msgnum;
	StrBuf *Caldata = NULL;
	char *euid;

	euid = strrchr(ChrPtr(ThisHref), '/');
	if (euid != NULL) {
		++euid;
	}
	else {
		euid = (char *) ChrPtr(ThisHref);
	}

	char *unescaped_euid = strdup(euid);
	if (!unescaped_euid) {
		return;
	}
	unescape_input(unescaped_euid);

	msgnum = locate_message_by_uid(c, unescaped_euid);
	free(unescaped_euid);
	if (msgnum > 0) {
		Caldata = fetch_ical(c, msgnum);
	}
	else {
		Caldata = NULL;
	}

	cal_multiget_out(msgnum, ThisHref, Caldata, ReportOut, crp);

	if (Caldata != NULL) {
		FreeStrBuf(&Caldata);
	}
}


// Compare function for "time-range" tests (RFC4791 section 9.9)
// Returns nonzero if the supplied icalcomponent occurs within the specified time range
//
// IMPLEMENTATION NOTE:
// ical_ctdl_is_overlap() works because icaltime_compare() is really smart.
// It looks at the time zone of the dtstart/dtend and can apparently go back up the icalcomponent
// hierarchy to find its time zone data.  I tested this by creating an event with a fictional
// time zone and it did the right thing.  It even showed the fictional name to me.  This saves us
// from having to convert everything to UTC before comparing.  Nice!
//
int caldav_time_range_filter_matches(icalcomponent *cal, char *start_str, char *end_str) {

	struct icaltimetype search_start = icaltime_from_string(start_str);		// time range being searched
	struct icaltimetype search_end = icaltime_from_string(end_str);

	struct icaltimetype dtstart = icalcomponent_get_dtstart(cal);			// time of event
	struct icaltimetype dtend = icalcomponent_get_dtend(cal);
	if (icaltime_is_null_time(dtend)) {
		dtend = dtstart;
	}

	// If it is a recurring event, RRULE is available at this level.  We can handle it here.
	icalproperty *rrule = icalcomponent_get_first_property(cal, ICAL_RRULE_PROPERTY);
	if (rrule) {
		struct icaldurationtype dur = icaltime_subtract(dtend, dtstart);	// recurrences need duration to find dtend
		struct icalrecurrencetype recur = icalproperty_get_rrule(rrule);
		icalrecur_iterator *ritr = icalrecur_iterator_new(recur, dtstart);	// iterate through recurrences
		while (dtstart = icalrecur_iterator_next(ritr), !icaltime_is_null_time(dtstart)) {
			dtend = icaltime_add(dtstart, dur);

			// Does THIS recurrence match the query?  If so, free the memory we used and stop iterating.
			if (ical_ctdl_is_overlap(dtstart, dtend, search_start, search_end)) {
				icalrecur_iterator_free(ritr);
				return(1);
			}
		}

		icalrecur_iterator_free(ritr);
		return(0);				// compared all recurrences, no match was found for any of them
	}

	// For non recurring events, do a simple time range compare.
	return(ical_ctdl_is_overlap(dtstart, dtend, search_start, search_end));	// We have a convenience function for this.
}


// Recursive function to apply CalDAV FILTERS to a calendar item.
// Returns zero if the calendar item was disqualified by a filter, nonzero if the calendar item still qualifies.
int caldav_apply_filters(void *cal, Array *filters, int apply_at_level) {

	int f = 0;						// filter number iterator
	int qual = 1;						// 0 for disqualify, 1 for qualify
	int previous_level = -1;
	int disregard_further_comp_filters = 0;

	while ( (f<array_len(filters)) && (qual) ) {

		// Tokenize the filter (a future performance hack would be to pre-tokenize instead of storing delimited strings)
		char this_filter[SIZ];
		safestrncpy(this_filter, array_get_element_at(filters, f), sizeof(this_filter));
		char *t[10] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL } ;
		char *ft = this_filter;
		int num_tokens = 0;
		while ( (t[num_tokens]=strtok_r(ft, "|", &ft)) && (num_tokens<10) ) {
			++num_tokens;
		}
		int this_rule_level = atoi(t[0]);
		// syslog(LOG_DEBUG, "caldav_apply_filters() filter=%d, level=%d, <%s>", f, this_rule_level, array_get_element_at(filters, f) );

		// Handle the individual filters defined in RFC4791 9.7.1 through 9.7.5

		if (apply_at_level < previous_level) {
			// syslog(LOG_DEBUG, "caldav: walking back down");
			return(qual);
		}

		else if (this_rule_level != apply_at_level) {
			// syslog(LOG_DEBUG, "caldav: apply_at_level=%d, this_rule_level=%d, skipping this rule", apply_at_level, this_rule_level);
		}

		else if (	(!strcasecmp(t[1], "comp-filter"))		// RFC4791 9.7.1 - filter by component
				&& (!disregard_further_comp_filters)		// one is enough to succeed
			) {
			// syslog(LOG_DEBUG, "component filter at level %d", this_rule_level);

			// comp-filter requires exactly one parameter (name="VXXXX")
			if (num_tokens < 4) {
				syslog(LOG_DEBUG, "caldav: comp-filter has no parameters - rejecting");
				return(0);
			}

			// Root element is NOT a component, but the root filter is "comp-filter" -- reject!
			if ( (!icalcomponent_isa_component(cal)) && (this_rule_level == 0) ) {
				syslog(LOG_DEBUG, "caldav: root element is not a component, rejecting");
				return(0);
			}

			// Current element is a component and the filter is "comp-filter" -- see if it matches the requested type
			if (	(icalcomponent_isa_component(cal))
				&& (!strcasecmp(t[2], "name"))
			) {
				if (icalcomponent_isa(cal) == icalcomponent_string_to_kind(t[3]) ) {
					// syslog(LOG_DEBUG, "caldav: component at level %d is <%s>, looking for <%s>, recursing...",
						//apply_at_level,
						//icalcomponent_kind_to_string(icalcomponent_isa(cal)), t[3]
					//);

					// We have a match.  Drill down into the subcomponents.

					icalcomponent *c = NULL;
					int number_of_subcomponents = 0;
					int number_of_matches = 0;
					for (	c = icalcomponent_get_first_component(cal, ICAL_ANY_COMPONENT);
						(c != 0);                                                
						c = icalcomponent_get_next_component(cal, ICAL_ANY_COMPONENT)
					) {
						++number_of_subcomponents;
						if (caldav_apply_filters(c, filters, apply_at_level+1)) {
							syslog(LOG_DEBUG, "Subcomponent %d might match", number_of_subcomponents);
							++number_of_matches;
						}
					}
					if (number_of_matches > 0) {			// something matched
						qual = 1;
						disregard_further_comp_filters = 1;

					}
					else if (number_of_subcomponents > 0) {		// nothing matched
						return(0);				// but only fail if there *were* subcomponents.
					}

				}
				else {
					//syslog(LOG_DEBUG, "caldav: component at level %d is <%s>, looking for <%s>, rejecting",
						//apply_at_level,
						//icalcomponent_kind_to_string(icalcomponent_isa(cal)),
						//t[3]
					//);
					return(0);
				}
			}

		}

		else if (!strcasecmp(t[1], "prop-filter")) {			// RFC4791 9.7.2 - filter by property
			syslog(LOG_DEBUG, "property filter at level %d FIXME not implemented yet", this_rule_level);
		}

		else if (!strcasecmp(t[1], "param-filter")) {			// RFC4791 9.7.3 - filter by parameter
			syslog(LOG_DEBUG, "parameter filter at level %d FIXME not implemented yet", this_rule_level);
		}

		else if (!strcasecmp(t[1], "is-not-defined")) {			// RFC4791 9.7.4
			syslog(LOG_DEBUG, "is-not-defined filter at level %d FIXME not implemented yet", this_rule_level);
		}

		else if (!strcasecmp(t[1], "text-match")) {			// RFC4791 9.7.5
			syslog(LOG_DEBUG, "text match filter at level %d FIXME not implemented yet", this_rule_level);
		}

		else if (!strcasecmp(t[1], "time-range")) {			// RFC4791 9.9
			syslog(LOG_DEBUG, "time range filter at level %d", this_rule_level);
			char *tr_start	= (char *)the_beginning_of_time;	// default if not specified
			char *tr_end	= (char *)the_end_of_time;		// default if not specified
			for (int i=2; (i+1)<num_tokens; i+=2) {
				if (!strcasecmp(t[i], "start")) {
					tr_start = t[i+1];
				}
				else if (!strcasecmp(t[i], "end")) {
					tr_end = t[i+1];
				}
			}
			if (caldav_time_range_filter_matches(cal, tr_start, tr_end)) {
				syslog(LOG_DEBUG, "time range matches");
			}
			else {
				syslog(LOG_DEBUG, "time range does not match -- rejecting");
				qual = 0;
			}
		}

		++f;
	}

	// syslog(LOG_DEBUG, "caldav: we reached the end of level %d , returning %d", apply_at_level, qual);
	return(qual);
}


// Called by report_the_room_itself() in room_functions.c when a CalDAV REPORT method
// is requested on a calendar room.  We fire up an XML Parser to decode the request and
// hopefully produce the correct output.
void caldav_report(struct http_transaction *h, struct ctdlsession *c) {
	struct cr_params crp;
	char buf[1024];

	memset(&crp, 0, sizeof(struct cr_params));

	XML_Parser xp = XML_ParserCreateNS("UTF-8", ':');
	if (xp == NULL) {
		syslog(LOG_INFO, "Cannot create XML parser!");
		do_404(h);
		return;
	}

	XML_SetElementHandler(xp, caldav_xml_start, caldav_xml_end);
	XML_SetCharacterDataHandler(xp, caldav_xml_chardata);
	XML_SetUserData(xp, &crp);
	XML_SetDefaultHandler(xp, NULL);	// Disable internal entity expansion to prevent "billion laughs attack"
	XML_Parse(xp, h->request_body, h->request_body_length, 1);
	XML_ParserFree(xp);

	if (crp.Chardata != NULL) {		// Discard any trailing chardata ... normally nothing here
		FreeStrBuf(&crp.Chardata);
		crp.Chardata = NULL;
	}

	// We're going to make a lot of MSG4 calls, and the preferred MIME type we want is "text/calendar".
	// The iCalendar standard is mature now, and we are no longer interested in text/x-vcal or application/ics.
	ctdl_printf(c, "MSGP text/calendar");
	ctdl_readline(c, buf, sizeof buf);

	// Now begin the REPORT.
	syslog(LOG_DEBUG, "CalDAV REPORT type is: %d", crp.report_type);
	StrBuf *ReportOut = NewStrBuf();
	StrBufAppendPrintf(ReportOut,
		"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		"<D:multistatus xmlns:D=\"DAV:\" xmlns:C=\"urn:ietf:params:xml:ns:caldav\">"
	);

	// RFC4791 7.8 "calendar-query" REPORT - Client will send a lot of search criteria.
	if (crp.report_type == cr_calendar_query) {
		int i = 0;
		Array *msglist = get_msglist(c, "ALL");
		if (msglist != NULL) {
			for (i = 0; i < array_len(msglist); ++i) {
				long m;
				memcpy(&m, array_get_element_at(msglist, i), sizeof(long));

				// load and parse one calendar item
				StrBuf *one_item = fetch_ical(c, m);
				icalcomponent *cal = icalcomponent_new_from_string(ChrPtr(one_item));

				// Does this calendar item qualify for output?  Run this calendar item through the filters.
				// syslog(LOG_DEBUG, "Evaluating message \033[33m%ld\033[0m...", m);
				if (caldav_apply_filters(cal, crp.filters, 0)) {
					syslog(LOG_DEBUG, "Message %ld \033[32mQUALIFIES\033[0m", m);

					// FIXME need to populate the Href instead of NULL
					StrBuf *FIXME = NewStrBufPlain(HKEY("https://FIX.ME.COM/WOW/EEK"));
					cal_multiget_out(m, FIXME, one_item, ReportOut, &crp);

				}
				else {
					syslog(LOG_DEBUG, "Message %ld \033[31mDOES NOT QUALIFY\033[0m", m);
				}
				// syslog(LOG_DEBUG, "");

				icalcomponent_free(cal);
				FreeStrBuf(&one_item);

			}
			array_free(msglist);
		}
	}

	// RFC4791 7.9 "calendar-multiget" REPORT - go get the specific Hrefs the client asked for.
	// Can we move this back into citserver too?
	else if ( (crp.report_type == cr_calendar_multiget) && (crp.Hrefs != NULL) ) {

		StrBuf *ThisHref = NewStrBuf();
		const char *pvset = NULL;
		while (StrBufExtract_NextToken(ThisHref, crp.Hrefs, &pvset, '|') >= 0) {
			StrBufTrim(ThisHref);				// remove leading/trailing whitespace from the href
			caldav_report_one_item(h, c, ReportOut, ThisHref, &crp);
		}
		FreeStrBuf(&ThisHref);
	}

	// RFC4791 7.10 "free-busy-query" REPORT
	else if (crp.report_type == cr_freebusy_query) {
		// FIXME build this REPORT.  At the moment we send an empty multistatus.
	}

	// Free any query parameters that might have been allocated during the xml parse
	if (crp.Hrefs != NULL) {
		FreeStrBuf(&crp.Hrefs);
		crp.Hrefs = NULL;
	}
	if (crp.filters) {
		array_free(crp.filters);
		crp.filters = NULL;
	}

	StrBufAppendPrintf(ReportOut, "</D:multistatus>\n");		// End the REPORT.

	add_response_header(h, strdup("Content-type"), strdup("text/xml"));
	h->response_code = 207;
	h->response_string = strdup("Multi-Status");
	h->response_body_length = StrLength(ReportOut);
	h->response_body = SmashStrBuf(&ReportOut);
}
