// webcit.h - "header of headers"
//
// Copyright (c) 1996-2024 by the citadel.org team
//
// This program is open source software.  Use, duplication, or
// disclosure is subject to the GNU General Public License v3.

// uncomment one or more of these to see raw http transactions
//#define DEBUG_HTTP
//#define REQUEST_BODY_TO_STDERR
//#define RESPONSE_BODY_TO_STDERR
//#define DEBUG_XML_PARSE

#define SHOW_ME_VAPPEND_PRINTF

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/un.h>
#include <sys/poll.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <stdarg.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <limits.h>
#include <iconv.h>
#include <libical/ical.h>
#include <libcitadel.h>
#define OPENSSL_NO_KRB5				// Work around RedHat's b0rken OpenSSL includes
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <expat.h>
#define _(x)	x				// temporary hack until we add i18n back in

// XML_StopParser is present in expat 2.x
#if XML_MAJOR_VERSION < 2
#error WebCit requires expat v2.0 or newer.
#endif

struct client_handle {				// this gets passed up the stack from the webserver to the application code
	int sock;
	SSL *ssl_handle;
};

struct keyval {					// key/value pair (for array)
	char *key;
	char *val;
};

struct http_transaction {			// The lifetime of an HTTP request goes through this data structure.
	char *method;				// The top half is built up by the web server and sent up to the
	char *url;				// application stack.  The second half is built up by the application
	char *http_version;			// stack and sent back down to the web server, which transmits it to
	char *site_prefix;			// the client.
	Array *request_headers;
	char *request_body_with_synth_headers;	// This is the request body with some synthetic headers prepended into it.
	char *request_body;			// this is just going to be a pointer into request_body_with_synth_headers
	long request_body_length;
	Array *request_parms;			// anything after the "?" in the URL
	int response_code;
	char *response_string;
	Array *response_headers;
	char *response_body;
	long response_body_length;
};

#define AUTH_MAX 256				// Maximum length of an HTTP AUTH header or equivalent cookie data
struct ctdlsession {
	struct ctdlsession *next;
	int is_bound;				// Nonzero if this record is currently bound to a running thread
	int sock;				// Socket connection to Citadel server
	char auth[AUTH_MAX];			// Auth string (empty if not logged in)
	char whoami[64];			// Display name of currently logged in user (empty if not logged in)
	char room[128];				// What room we are currently in
	int room_current_view;
	int room_default_view;
	int is_trash_folder;			// nonzero if this room is the user's Trash folder
	int is_room_aide;			// nonzero if the user has aide rights to THIS room
	int can_delete_messages;		// nonzero if the user is permitted to delete messages in THIS room
	long last_seen;
	int new_messages;
	int total_messages;
	time_t last_access;			// Timestamp of last request that used this session
	time_t num_requests_handled;
	time_t room_mtime;			// Timestampt of the most recent write activity in this room
};

struct uploaded_file {				// things that have been uploaded to the server (such as email attachments)
	char id[10];
	char filename[256];
	char content_type[256];
	long length;
	FILE *fp;
};

extern char *ssl_cipher_list;
extern int is_https;				// nonzero if we are an HTTPS server today
extern char *ctdl_dir;				// directory where Citadel Server is running
void init_ssl(void);
void starttls(struct client_handle *);
void endtls(struct client_handle *);
int client_write_ssl(struct client_handle *ch, char *buf, int nbytes);
int client_read_ssl(struct client_handle *ch, char *buf, int nbytes);

enum {
	WEBSERVER_HTTP,
	WEBSERVER_HTTPS,
	WEBSERVER_UDS
};

#define TRACE syslog(LOG_DEBUG, "\033[3%dmCHECKPOINT: %s:%d\033[0m", ((__LINE__%6)+1), __FILE__, __LINE__)
#define SLEEPING		180		// TCP connection timeout
#define MAX_WORKER_THREADS	32		// Maximum number of worker threads permitted to exist
#define DEFAULT_SSL_CIPHER_LIST "DEFAULT"       // See http://openssl.org/docs/apps/ciphers.html
#define WEBSERVER_PORT		80
#define WEBSERVER_INTERFACE	"*"
#define CTDL_DIR		"/usr/local/citadel"
#define DEVELOPER_ID		0
#define CLIENT_ID		4
#define TARGET			"webcit02"	// Window target for inline URL's
#define ROOMNAMELEN		128		// The size of a roomname string
#define	DAV_MOVE		0		// MOVE=0 COPY=1 don't change these!
#define DAV_COPY		1		// they are the values used in the Citadel Server MOVE command

// FIXME retrieve this from the server
#define default_zone_name	""

// admin_functions.c
void try_login(struct http_transaction *, struct ctdlsession *);
void logout(struct http_transaction *, struct ctdlsession *);
void whoami(struct http_transaction *, struct ctdlsession *);
void biff(struct http_transaction *, struct ctdlsession *);
void ctdl_a(struct http_transaction *, struct ctdlsession *);

// caldav_reports.c
void caldav_xml_start(void *, const char *, const char **);
void caldav_xml_end(void *, const char *);
void caldav_xml_chardata(void *, const XML_Char *, int);
StrBuf *fetch_ical(struct ctdlsession *, long);
int caldav_time_range_filter_matches(icalcomponent *, char *, char *);
int caldav_apply_filters(void *, Array *, int);
void caldav_report(struct http_transaction *, struct ctdlsession *);

// ctdlclient.c
int ctdl_read_binary(struct ctdlsession *, char *, int);
int ctdl_readline(struct ctdlsession *, char *, int);
StrBuf *ctdl_readtextmsg(struct ctdlsession *);
ssize_t ctdl_write(struct ctdlsession *, const void *, size_t);
void ctdl_printf(struct ctdlsession *, const char *, ...);
int uds_connectsock(char *);
void extract_auth(struct http_transaction *, char *, int);
int login_to_citadel(struct ctdlsession *, char *, char *);
struct ctdlsession *connect_to_citadel(struct http_transaction *);
void disconnect_from_citadel(struct ctdlsession *);

// ctdl_commands.c
void serv_info(struct http_transaction *, struct ctdlsession *);
void ctdl_c(struct http_transaction *, struct ctdlsession *);

// ctdlfunctions.c
void ctdl_delete_msgs(struct ctdlsession *, long *, int);

// floor_functions.c
void floor_list(struct http_transaction *, struct ctdlsession *);
void ctdl_f(struct http_transaction *, struct ctdlsession *);

// forum_view.c
void setup_for_forum_view(struct ctdlsession *);
JsonValue *json_tokenize_recipients(const char *, long, char *);
void json_render_one_message(struct http_transaction *, struct ctdlsession *, long);

// html2html.c
void stripquotes(char *);
void extract_charset_from_meta(char *, char *, char *);
StrBuf *html2html(const char *, int, char *, long, StrBuf *);
void UrlizeText(StrBuf *, StrBuf *, StrBuf *);
void url(char *, size_t);

// http.c
int client_write(struct client_handle *, char *, int);
int client_read(struct client_handle *, char *, int);
int client_readline(struct client_handle *, char *, int);
void client_printf(struct client_handle *, const char *, ...);
void add_response_header(struct http_transaction *, char *, char *);
void perform_one_http_transaction(struct client_handle *);
char *header_val(struct http_transaction *, char *);
char *get_url_param(struct http_transaction *, char *);

// ical_ctdl_is_overlap.c
int ical_ctdl_is_overlap(struct icaltimetype, struct icaltimetype, struct icaltimetype, struct icaltimetype);

// main.c
int main(int, char **);

// messages.c
long locate_message_by_uid(struct ctdlsession *, char *);
void dav_delete_message(struct http_transaction *, struct ctdlsession *, long);
void dav_move_or_copy_message(struct http_transaction *, struct ctdlsession *, long, int);
void dav_get_message(struct http_transaction *, struct ctdlsession *, long);
void dav_put_message(struct http_transaction *, struct ctdlsession *, char *, long);
void download_mime_component(struct http_transaction *, struct ctdlsession *, long, char *);

// request.c
void do_404(struct http_transaction *);
void do_405(struct http_transaction *);
void do_412(struct http_transaction *);
void do_204(struct http_transaction *);
void do_502(struct http_transaction *);
void request_http_authenticate(struct http_transaction *);
void http_redirect(struct http_transaction *, char *);
void perform_request(struct http_transaction *);

// room_functions.c
Array *get_msglist(struct ctdlsession *, char *);
int match_etags(char *, long);
void json_stat(struct http_transaction *, struct ctdlsession *);
void json_mailbox(struct http_transaction *, struct ctdlsession *);
void json_msglist(struct http_transaction *, struct ctdlsession *, char *);
void read_room_info_banner(struct http_transaction *, struct ctdlsession *);
void set_last_read_pointer(struct http_transaction *, struct ctdlsession *);
void object_in_room(struct http_transaction *, struct ctdlsession *);
void report_the_room_itself(struct http_transaction *, struct ctdlsession *);
void options_the_room_itself(struct http_transaction *, struct ctdlsession *);
void propfind_the_room_itself(struct http_transaction *, struct ctdlsession *);
void get_the_room_itself(struct http_transaction *, struct ctdlsession *);
void the_room_itself(struct http_transaction *, struct ctdlsession *);
void room_list(struct http_transaction *, struct ctdlsession *);
void ctdl_r(struct http_transaction *, struct ctdlsession *);

// static.c
void output_static(struct http_transaction *);

// tcp_sockets.c
int lingering_close(int);
int webcit_tcp_server(const char *, int, int);
int webcit_uds_server(char *, int);

// text2html.c
StrBuf *text2html(const char *, int, char *, long, StrBuf *);
StrBuf *variformat2html(StrBuf *);

// tls.c
void bind_to_key_and_certificate(void);
void init_ssl(void);
void update_key_and_cert_if_needed(void);
void starttls(struct client_handle *);
void endtls(struct client_handle *);
int client_write_ssl(struct client_handle *, char *, int);
int client_read_ssl(struct client_handle *, char *, int);

// upload.c
void upload_handler(char *, char *, char *, char *, void *, char *, char *, size_t, char *, char *, void *);
void upload_files(struct http_transaction *, struct ctdlsession *);
void ctdl_p_base(struct http_transaction *, struct ctdlsession *);
void delete_upload(struct uploaded_file);
void dav_delete_upload(struct http_transaction *, struct ctdlsession *, struct uploaded_file);
struct uploaded_file pop_upload(char *);
void attachment_filter(char *, char *, char *, char *, void *, char *, char *, size_t, char *, char *, void *);
void load_attachments_from_message(struct http_transaction *, struct ctdlsession *, char *);
void specific_upload(struct http_transaction *, struct ctdlsession *, char *);
void ctdl_p(struct http_transaction *, struct ctdlsession *);

// user_functions.c
void fetch_user_photo(struct http_transaction *, struct ctdlsession *, char *);
void fetch_user_bio(struct http_transaction *, struct ctdlsession *, char *);
void object_in_user(struct http_transaction *, struct ctdlsession *, char *);
void the_user_itself(struct http_transaction *, struct ctdlsession *, char *);
void user_list(struct http_transaction *, struct ctdlsession *);
void ctdl_u(struct http_transaction *, struct ctdlsession *);

// util.c
int unescape_input(char *);
char *http_datestring(time_t);

// webserver.c
void spawn_another_worker_thread(int *);
void worker_entry(int *);
int webserver(char *, int, int);

